#ifndef FrameCPP_VERSION_8_FrVect_HH
#define FrameCPP_VERSION_8_FrVect_HH

#include <memory>
#include <vector>
#include <string>

#include "general/types.hh"
#include "framecpp/FrameCPP.hh"

#include "framecpp/Common/Compression.hh"
#include "framecpp/Common/FrVect.hh"

#include "framecpp/Version7/FrVect.hh"

#include "framecpp/Version8/FrameSpec.hh"
#include "framecpp/Version8/Dimension.hh"
#include "framecpp/Version8/GPSTime.hh"
#include "framecpp/Version8/STRING.hh"

namespace FrameCPP
{
  namespace Version_8
  {
    using Previous::FrVectDataTypes;

    //===================================================================
    /// \brief Vector Data Structure Definition
    //===================================================================
    class FrVect
      : public ObjectWithChecksum< Common::FrameSpec::Object >,
        public Common::FrVect,
        public FrVectDataTypes
    {
    public:
      enum compression_scheme_type {
	//---------------------------------------------------------
	//---------------------------------------------------------

#if WORDS_BIGENDIAN
#define NATIVE_DECL(x) x = BIGENDIAN_##x
#else
#define NATIVE_DECL(x) x = LITTLEENDIAN_##x
#endif
#define CST_DECL(x,y) BIGENDIAN_##x = y, LITTLEENDIAN_##x = y + 256, NATIVE_DECL(x)

	CST_DECL(RAW,0),
	CST_DECL(GZIP,1),
        CST_DECL(DIFF_GZIP,3),
	CST_DECL(ZERO_SUPPRESS_WORD_2,5),
	CST_DECL(ZERO_SUPPRESS_WORD_4,8),
	CST_DECL(ZERO_SUPPRESS_WORD_8,10),

#undef CST_DECL
#undef NATIVE_DECL

	//---------------------------------------------------------
	// Declaration of meta modes
	//---------------------------------------------------------
	ZERO_SUPPRESS_OTHERWISE_GZIP = 6,
	BEST_COMPRESSION = 9,

	//---------------------------------------------------------
	// Backwards compatability
	//---------------------------------------------------------
	ZERO_SUPPRESS_SHORT = ZERO_SUPPRESS_WORD_2,
	ZERO_SUPPRESS_INT_FLOAT = ZERO_SUPPRESS_WORD_4,
#undef GENERIC
	NONE = RAW
      };

      enum data_types_type {
	FR_VECT_C = 0,
	FR_VECT_2S = 1,
	FR_VECT_8R = 2,
	FR_VECT_4R = 3,
	FR_VECT_4S = 4,
	FR_VECT_8S = 5,
	FR_VECT_8C = 6,
	FR_VECT_16C = 7,
	FR_VECT_STRING = 8,
	FR_VECT_2U = 9,
	FR_VECT_4U = 10,
	FR_VECT_8U = 11,
	FR_VECT_1U = 12
      };


      typedef INT_2U compress_type;
      typedef INT_2U type_type;
      typedef INT_4U nData_type;
      typedef INT_4U nBytes_type;
      typedef INT_4U nDim_type;

      static const int DEFAULT_GZIP_LEVEL;

      //-----------------------------------------------------------------
      /// \brief Promote previous version of object to this version
      ///
      /// \param[in] Source
      ///     The frame specification version of the object.
      /// \param[in] Obj
      ///     The previous version of the object.
      /// \param[in] Stream
      ///     Stream from which the origional object was created.
      ///
      /// \return
      ///     A new instance of FrameH based on Obj
      //-----------------------------------------------------------------
      static std::auto_ptr< Object >
      Promote( INT_2U Source,
	       Object* Obj, Common::IStream* Stream );

      //-----------------------------------------------------------------
      /// \brief The name structure name of this object.
      ///
      /// \return
      ///     The name of the structure as specified by the frame
      ///     specification.
      //-----------------------------------------------------------------
      static const char* const StructName( );

      //-----------------------------------------------------------------
      /// \brief The description of structure
      ///
      /// \return
      ///     A Description object which describes this structure as
      ///     specified by the frame specification.
      //-----------------------------------------------------------------
      static const Common::Description* StructDescription( );

      //-----------------------------------------------------------------
      /// \brief Default constructor
      ///
      /// \return
      ///    A new instance of this object.
      //-----------------------------------------------------------------
      FrVect( );

      //-----------------------------------------------------------------
      /// \brief Copy Constructor
      ///
      /// \param[in] vect
      ///     The object from which to copy the information.
      ///
      /// \return
      ///    A new instance of this object.
      //-----------------------------------------------------------------
      FrVect( const FrVect& vect );


      //-----------------------------------------------------------------
      /// \brief Constructor.
      ///
      /// \param[in] name
      ///     The name of the data.
      /// \param[in] type
      ///     The data type.
      /// \param[in] nDim
      ///     The number of dimensions.
      /// \param[in] dims
      ///     A pointer to 'nDim' Dimension objects containing
      ///     information about the dimensions for this data.
      /// \param[in] byte_order
      ///     Byte order of the data. Default is BYTE_ORDER_HOST
      /// \param[in] data
      ///     A pointer to the data.  Default: 0
      /// \param[in] unitY
      ///     Units for the data.  Default: ""
      ///
      /// \return
      ///    A new instance of this object.
      //-----------------------------------------------------------------
      FrVect( const std::string& name, INT_2U type, INT_4U nDim,
	      const Dimension* dims,
	      const byte_order_type byte_order = BYTE_ORDER_HOST,
	      const void* data = 0, 
	      const std::string& unitY = "" );

      //-----------------------------------------------------------------
      /// \brief Constructor.
      ///
      ///
      /// \param[in] name
      ///     The name of the data.
      /// \param[in] type
      ///     The data type.
      /// \param[in] nDim
      ///     The number of dimensions.
      /// \param[in] dims
      ///     A pointer to 'nDim' Dimension objects containing
      ///     information about the dimensions for this data.
      /// \param[in] byte_order
      ///     Byte order of the data. Default is BYTE_ORDER_HOST.
      /// \param[in] data
      ///     A pointer to the data.  Default: 0
      /// \param[in] unitY
      ///     Units for the data.  Default: ""
      /// \param[in] allocate
      ///     Whether to allocate data. Default: true.
      /// \param[in] owns
      ///     Whether to destruct data in the destructor.
      ///     Default: true.
      /// \return
      ///    A new instance of this object.
      //-----------------------------------------------------------------
      FrVect( const std::string& name, INT_2U type, INT_4U nDim,
	      const Dimension* dims,
	      const byte_order_type byte_order = BYTE_ORDER_HOST,
	      void* data = 0,
	      const std::string& unitY = "", 
	      bool allocate = true, bool owns = true );

      FrVect( const std::string& name,
	      compress_type Compress,
	      type_type type,
	      nDim_type nDim,
	      const Dimension* dims,
	      nData_type NData,
	      nBytes_type NBytes,
	      void* data,
	      const std::string& unitY = "", 
	      bool allocate = true,
	      bool owns = true );

      template< class T >
      FrVect( const std::string& name, INT_4U nDim = 1,
	      const Dimension* dims = 0,
	      const T* data = 0, const std::string& unitY = "" );

      template< class T >
      FrVect( const std::string& name, INT_4U nDim = 1,
	      const Dimension* dims = 0,
	      T* data = 0, const std::string& unitY = "",
	      bool allocate = true, bool owns = true );

      //-----------------------------------------------------------------
      /// \brief Promotion Constructor
      ///
      /// \param[in] Source
      ///     An FrAdcData structure from a previous frame specification.
      /// \param[in] Stream
      ///     The stream from which the earliest version of the
      ///     FrAdcData structure was read.
      ///
      /// \return
      ///     A new instance of the object.
      //-----------------------------------------------------------------
      FrVect( const Previous::FrVect& Source,
	      Common::IStream* Stream );

      //-----------------------------------------------------------------
      /// \brief Destructor
      //-----------------------------------------------------------------
      virtual ~FrVect( );

      //-----------------------------------------------------------------
      /// \brief The name structure name of this object.
      ///
      /// \return
      ///     The name of the structure as specified by the frame
      ///     specification.
      //-----------------------------------------------------------------
      virtual const char* const ObjectStructName( ) const;

      //-----------------------------------------------------------------
      /// \brief equality operator
      ///
      /// \param[in] RHS
      ///     The FrAdcData object to be compared.
      ///
      /// \return
      ///     The value true is returned if this object is equivelent
      ///     to the RHS instance; false otherwise.
      //-----------------------------------------------------------------
      bool operator==( const FrVect& RHS ) const;

      //-----------------------------------------------------------------
      /// \brief inequality operator
      ///
      /// \param[in] RHS
      ///     The FrAdcData object to be compared.
      ///
      /// \return
      ///     The value true is returned if this object is not equivelent
      ///     to the RHS instance; false otherwise.
      //-----------------------------------------------------------------
      bool operator!=( const FrVect& RHS ) const;

      //-----------------------------------------------------------------
      /// \brief Concatinate with another FrAdcData
      ///
      /// \param[in] RHS
      ///     The source of the information to append to this FrAdcData
      ///     structure.
      ///
      /// \return
      ///     A reference to this object
      //-----------------------------------------------------------------
      FrVect& operator+=( const FrVect& RHS );

      //-----------------------------------------------------------------
      /// \brief comparison operator
      //-----------------------------------------------------------------
      virtual bool operator==( const Common::FrameSpec::Object& Obj ) const;

      //-----------------------------------------------------------------
      /// \brief Retrieve the channel name.
      ///
      /// \return
      ///     The channel name
      //-----------------------------------------------------------------
      const std::string& GetName( ) const;

      //-----------------------------------------------------------------
      /// \brief Retrieve the compression algorithm number.
      ///
      /// \return
      ///     The compression algorithm number.
      //-----------------------------------------------------------------
      INT_2U GetCompress( ) const;

      //-----------------------------------------------------------------
      /// \brief Retrieve the vector class.
      ///
      /// \return
      ///     The vector class.
      //-----------------------------------------------------------------
      INT_2U GetType( ) const;

      //-----------------------------------------------------------------
      /// \brief Retrieve the number of sample elements in data series.
      ///
      /// \return
      ///     The number of sample elements in data series.
      //-----------------------------------------------------------------
      INT_8U GetNData( ) const;

      //-----------------------------------------------------------------
      /// \brief Retrieve the number of bytes in the compressed vector.
      ///
      /// \return
      ///     The number of bytes in the compressed vector.
      //-----------------------------------------------------------------
      INT_8U GetNBytes( ) const;

      //-----------------------------------------------------------------
      /// \brief Retrieve the pointer to the data.
      ///
      /// \return
      ///     The pointer to the data.
      ///
      /// \note
      ///     This will expand the data if it is currently compressed.
      //-----------------------------------------------------------------
      const CHAR_U* GetData( ) const;

      //-----------------------------------------------------------------
      /// \brief Retrieve the pointer to the compressed data.
      ///
      /// \return
      ///     The pointer to the compressed data.
      //-----------------------------------------------------------------
      const CHAR_U* GetDataRaw( ) const;

      //-----------------------------------------------------------------
      /// \brief Retrieve the pointer to the compressed data.
      ///
      /// \return
      ///     The pointer to the compressed data.
      //-----------------------------------------------------------------
      CHAR_U* GetDataRaw( );

      //-----------------------------------------------------------------
      /// \brief Retrieve the pointer to the uncompressed data.
      ///
      /// \return
      ///     The pointer to the uncompressed data.
      //-----------------------------------------------------------------
      CHAR_U* GetDataUncompressed( );

      //-----------------------------------------------------------------
      /// \brief Retrieve the pointer to the uncompressed data.
      ///
      /// \return
      ///     The pointer to the uncompressed data.
      //-----------------------------------------------------------------
      const CHAR_U* GetDataUncompressed( General::AutoArray< CHAR_U >&
					 Expanded ) const;

      template< class T > static INT_2U GetDataType();

      INT_4U GetNDim( ) const;

      const Dimension& GetDim( INT_4U Offset ) const;

      Dimension& GetDim( INT_4U Offset );

      const std::string GetUnitY( ) const;

      CHAR_U* SteelData( bool& Owns ) const;

      void Compress( compression_scheme_type Scheme, int GZipLevel );
      void Uncompress( );

      virtual void CompressData( INT_4U Scheme, INT_2U GZipLevel );

      virtual Compression::compression_base_type
      Compression( ) const;

      virtual Common::FrameSpec::Object*
      CloneCompressed( cmn_compression_scheme_type Scheme,
		       cmn_compression_level_type Level ) const;

      //-----------------------------------------------------------------
      /// \brief Establish the channel name.
      ///
      /// \param[in] Name
      ///     The channel name
      //-----------------------------------------------------------------
      void SetName( const std::string& Name );

      void SetNData( INT_4U NData );

      size_t GetTypeSize( ) const;
      static size_t GetTypeSize( INT_2U type );

      //-----------------------------------------------------------------
      /// \brief Merge with another FrAdcData
      ///
      /// \param[in] RHS
      ///     The source of the information to append to this FrAdcData
      ///     structure.
      ///
      /// \return
      ///     A reference to this object
      //-----------------------------------------------------------------
      FrVect& Merge( const FrVect& RHS );

      //-----------------------------------------------------------------
      /// \brief Request a subset of the data.
      ///
      /// \param[in] Start
      ///     The offset in the data set where to start.
      /// \param[in] Stop
      ///     The offset in the data set where to stop.
      ///
      /// \return
      ///     The subset of data bounded by data[Start, Stop).
      //-----------------------------------------------------------------
      std::auto_ptr< FrVect > SubFrVect( INT_4U Start, INT_4U Stop ) const;

    protected:
      //-----------------------------------------------------------------
      /// \brief Verify the integrity of the object
      ///
      /// \param Verifier
      ///     The handle to the Verify object used to store data that
      ///     needs to be accumulated before being processed.
      /// \param Stream
      ///     The input stream from which the object was created.
      ///
      /// \remarks
      ///     Upon failure, this method thows an exception.
      //-----------------------------------------------------------------
      virtual void VerifyObject( Common::Verify& Verifier,
				 Common::IFrameStream& Stream ) const;

      //-----------------------------------------------------------------
      /// \brief Down grade an object
      ///
      /// \param[in] Target
      ///     The version of the frame specification to demote too.
      /// \param[in] Stream
      ///     The input stream from which the original object was read.
      ///
      /// \return
      ///     An object of the previous generation.
      //-----------------------------------------------------------------
      virtual std::auto_ptr< Object >
      demote( INT_2U Target, Common::IStream* Stream ) const;

      //-----------------------------------------------------------------
      /// \brief Upgrade an object
      ///
      /// \param[in] Target
      ///     The version of the frame specification to promote too.
      /// \param[in] Stream
      ///     The input stream from which the original object was read.
      ///
      /// \return
      ///     An object of this class.
      //-----------------------------------------------------------------
      virtual std::auto_ptr< Object >
      promote( INT_2U Target, Object* Obj, Common::IStream* Stream ) const;

      //-----------------------------------------------------------------
      /// \brief Number of bytes needed to write this structure
      ///
      /// \param[in] Stream
      ///     The stream from which to the object is being read or
      ///     written.
      ///
      /// \return
      ///     The number of bytes need to read or write this object.
      //-----------------------------------------------------------------
      virtual FrameCPP::cmn_streamsize_type
      pBytes( const Common::StreamBase& Stream ) const;

      //-----------------------------------------------------------------
      /// \brief Stream constructor
      ///
      /// \param[in] Stream
      ///     The input stream from where the object is being read.
      //-----------------------------------------------------------------
      virtual FrVect* pCreate( Common::IStream& Stream ) const;

      //-----------------------------------------------------------------
      /// \brief Write the structure to the stream
      ///
      /// \param[in] Stream
      ///     The output stream where the object is to be written.
      //-----------------------------------------------------------------
      virtual void pWrite( Common::OStream& Stream ) const;

      void
      expandToBuffer( General::AutoArray< CHAR_U >& Dest,
		      INT_8U& DestNBytes ) const;

    private:
      //-----------------------------------------------------------------
      /// \brief Object with checksum data.
      //-----------------------------------------------------------------
      typedef ObjectWithChecksum< Common::FrameSpec::Object > object_type;

      //-----------------------------------------------------------------
      /// \brief Dictionary class id for this structure.
      //-----------------------------------------------------------------
      static const Common::FrameSpec::Info::frame_object_types s_object_id
      = Common::FrameSpec::Info::FSI_FR_VECT;
      
      using Common::FrameSpec::Object::Create;

      struct data_definition_type
      {
      public:

	//---------------------------------------------------------------
	/// Channel name.
	///
	/// \note
	///     Not required to be unique.
	//---------------------------------------------------------------
	STRING				name;
	//---------------------------------------------------------------
	/// Compression algorithm number
	//---------------------------------------------------------------
	INT_2U				compress;
	//---------------------------------------------------------------
	/// Vector class
	//---------------------------------------------------------------
	INT_2U				type;
	//---------------------------------------------------------------
	/// Number of sample elements in data series.
	//---------------------------------------------------------------
	INT_8U				nData;
	//---------------------------------------------------------------
	/// Number of bytes in the compressed vector.
	//---------------------------------------------------------------
	INT_8U				nBytes;
	//---------------------------------------------------------------
	/// Compressed data series.
	//---------------------------------------------------------------
	CHAR_U*				data;
	//---------------------------------------------------------------
	/// Dimension data for data series.
	//---------------------------------------------------------------
	std::vector< Dimension >	dimension;
	//---------------------------------------------------------------
	/// String describing how to interpret the value of each
	/// element.
	//---------------------------------------------------------------
	STRING				unitY;

        data_definition_type();
        data_definition_type( const std::string& n, INT_2U t, 
			      const byte_order_type byte_order, INT_4U ndim,
			      const Dimension* dims, const std::string& unit );

	void copy_core( const data_definition_type& Source );

	bool operator==( const data_definition_type& RHS ) const;

      };
 
      data_definition_type	m_data;
      bool			m_owns;

      //-----------------------------------------------------------------
      /// \brief Stream Constructor
      ///
      /// \param[in] Stream
      ///     The stream from which the object is being read.
      ///
      /// \return
      ///    A new instance of this object.
      //-----------------------------------------------------------------
      FrVect( Common::IStream& Stream );

      void copy_data( CHAR_U* Data, INT_8U size,
		      bool allocate, bool own );

    };


    inline Compression::compression_base_type FrVect::
    Compression( ) const
    {
      return ( 0x0FF & m_data.compress );
    }

    inline std::auto_ptr< Common::FrameSpec::Object > FrVect::
    Promote( INT_2U Source,
	     Common::FrameSpec::Object* Obj, Common::IStream* Stream )
    {
      return Object::PromoteObject< Previous::FrVect, FrVect >
	( DATA_FORMAT_VERSION, Source, Obj, Stream );
    }

    inline const char* const FrVect::
    StructName( )
    {
      static const CHAR* class_name( "FrVect" );
      return class_name;
    }

    //-------------------------------------------------------------------
    /// Return a pointer to the data being managed by the FrVect
    /// structure.
    ///
    /// \see GetDataRaw, GetDataUncompressed
    //-------------------------------------------------------------------
    inline const CHAR_U* FrVect::
    GetData( ) const
    {
      return GetDataRaw( );
    }

    inline INT_2U FrVect::
    GetCompress( ) const
    {
      return m_data.compress;
    }

    inline INT_2U FrVect::
    GetType( ) const
    {
      return m_data.type;
    }

    inline size_t FrVect::
    GetTypeSize( ) const
    {
      return GetTypeSize( m_data.type );
    }

    inline INT_8U FrVect::
    GetNBytes( ) const
    {
      return m_data.nBytes;
    }

    inline INT_4U FrVect::GetNDim( ) const
    {
      return m_data.dimension.size( );
    }

    inline const Dimension& FrVect::
    GetDim( INT_4U Offset ) const
    {
      //:TODO: Throw exception if value out of range
      return m_data.dimension[ Offset ];
    }

    inline Dimension& FrVect::
    GetDim( INT_4U Offset )
    {
      //:TODO: Throw exception if value out of range
      return m_data.dimension[ Offset ];
    }

    inline const std::string FrVect::
    GetUnitY( ) const
    {
      return m_data.unitY;
    }

    //-------------------------------------------------------------------
    /// Assign a new value for the channel name.
    //-------------------------------------------------------------------
    inline void FrVect::
    SetName( const std::string& Name )
    {
      m_data.name = Name;
    }

    inline void FrVect::
    SetNData( INT_4U NData )
    {
      m_data.nData = NData;
    }
   
    inline CHAR_U* FrVect::
    SteelData( bool& Owns ) const
    {
      Owns = m_owns;
      const_cast< FrVect* >( this )->Uncompress( );
      const_cast< FrVect* >( this )->m_owns = false;
      return m_data.data;
    }
   
  } // namespace - Version_8
} // namespace FrameCPP

#if defined( __SUNPRO_CC ) && ( __SUNPRO_CC <= 0x550 )
#include "framecpp/Version8/FrVect.icc"
#endif /* */

#endif /* FrameCPP_VERSION_8_FrVect_HH */
